#
# Memotech Emulator
#

CFLAGS =	-g -Werror -I$(TH) -I$(DIRT) -DUNIX -DLINUX -DLSB_FIRST -DDEBUG -DBEMEMU -DHAVE_CFX2 -DHAVE_VGA
LFLAGS =	-g
X11 =		/usr/lib/X11
LIBS =		-lrt -lportaudio -lm
TH =		../../th
DIRT =		../../dirt
IDIR =		../../memu

.SUFFIXES:	.c .o

.c.o:
		@echo "  CC    $@"
		@$(CC) -c -o $@ $(CFLAGS) $<

#
all:		$(IDIR)/memu $(IDIR)/memu-fb $(IDIR)/memu-cfg $(IDIR)/memu-cfb

$(IDIR)/memu:		Z80.o diag.o common.o win_x.o roms.o mem.o vid.o kbd2.o tape.o joy_lin.o dart.o snd.o ctc.o monprom.o mon.o sdxfdc.o sid.o printer.o spec.o cpm.o dis.o ui.o cfx2.o vga.o vdeb.o memu.o $(TH)/th.a $(DIRT)/dirt.a
		@echo "  LD    $@"
		@$(CC) -o $@ $(LFLAGS) $+ -lncurses -L$(X11)/lib -lX11 $(LIBS)

$(IDIR)/memu-fb:	Z80.o diag.o common.o win_fb.o roms.o mem.o vid.o kbd2.o tape.o joy_lin.o dart.o snd.o ctc.o monprom.o mon.o sdxfdc.o sid.o printer.o spec.o cpm.o dis.o ui.o cfx2.o vga.o vdeb.o memu.o $(TH)/th.a $(DIRT)/dirt.a
		@echo "  LD    $@"
		@$(CC) -o $@ $(LFLAGS) $+ -lncurses $(LIBS)

$(IDIR)/memu-cfg:	Z80.o diag.o common.o win_x.o roms.o mem_cfg.o vid.o kbd_cfg.o tape.o joy_lin.o dart.o snd.o ctc.o monprom.o mon.o sdxfdc.o sid.o printer.o spec.o cpm.o dis.o ui.o cfx2.o vga.o vdeb.o config.o memu_cfg.o $(TH)/th.a $(DIRT)/dirt.a
		@echo "  LD    $@"
		@$(CC) -o $@ $(LFLAGS) $+ -lncurses -L/usr/X11R6/lib -lX11 $(LIBS)

$(IDIR)/memu-cfb:	Z80.o diag.o common.o win_fb.o roms.o mem_cfg.o vid.o kbd_cfg.o tape.o joy_lin.o dart.o snd.o ctc.o monprom.o mon.o sdxfdc.o sid.o printer.o spec.o cpm.o dis.o ui.o cfx2.o vga.o vdeb.o config.o memu_cfg.o $(TH)/th.a $(DIRT)/dirt.a
		@echo "  LD    $@"
		@$(CC) -o $@ $(LFLAGS) $+ -lncurses $(LIBS)

Z80.o:		Z80.c Z80.h Tables.h CodesCB.h CodesXCB.h CodesED.h CodesXX.h Codes.h

diag.o:		diag.c types.h diag.h vdeb.h

common.o:	common.c types.h diag.h common.h mem.h vid.h kbd.h tape.h joy.h mon.h sdxfdc.h sid.h printer.h ui.h vdeb.h

win_x.o:		win_x.c types.h common.h win.h

win_fb.o:	win_fb.c types.h common.h win.h

roms.o:		roms.c types.h roms.h

mem.o:		mem.c types.h common.h diag.h roms.h mem.h

mem_cfg.o:	mem.c types.h common.h diag.h roms.h mem.h
		@echo "  CC    $@"
		@$(CC) -c -o $@ $(CFLAGS) -DDYNAMIC_ROMS $<

vid.o:		vid.c types.h common.h diag.h mem.h vid.h kbd.h mon.h

kbd2.o:		kbd2.c types.h common.h diag.h win.h kbd.h vdeb.h

kbd_cfg.o:	kbd2.c types.h common.h diag.h win.h kbd.h
		@echo "  CC    $@"
		@$(CC) -c -o $@ $(CFLAGS) -DALT_KEYPRESS=test_cfg_key $<

tape.o:		tape.c types.h memu.h tape.h

joy_lin.o:		joy_lin.c types.h common.h diag.h kbd.h joy.h

dart.o:		dart.c types.h common.h diag.h dart.h

snd.o:		snd.c types.h common.h diag.h snd.h

ctc.o:		ctc.c types.h common.h diag.h ctc.h

monprom.o:	monprom.c types.h monprom.h

mon.o:		mon.c types.h common.h diag.h win.h kbd.h monprom.h mon.h $(TH)/th.h

sdxfdc.o:	sdxfdc.c types.h diag.h sdxfdc.h

sid.o:		sid.c types.h diag.h common.h sid.h

printer.o:	printer.c types.h common.h diag.h printer.h

spec.o:		spec.c types.h diag.h common.h spec.h

cpm.o:		cpm.c Z80.h types.h common.h diag.h mem.h mon.h printer.h $(DIRT)/dirt.h

dis.o:		dis.c types.h mem.h dis.h

ui.o:		ui.c types.h common.h win.h mem.h vid.h monprom.h dis.h ui.h

cfx2.o:		cfx2.c types.h common.h diag.h cfx2.h

vga.o:		vga.c vga.h monprom.h win.h vid.h kbd.h common.h diag.h

vdeb.o:		vdeb.c vdeb.h memu.h mem.h dis.h kbd.h monprom.h win.h Z80.h

config.o:	config.c config.h types.h memu.h monprom.h win.h kbd.h memu.h monprom.h win.h kbd.h mem.h vid.h mon.h snd.h sdxfdc.h common.h
		@echo "  CC    $@"
		@$(CC) -c -o $@ $(CFLAGS) -DALT_MAIN $<

memu.o:		memu.c Z80.h types.h common.h diag.h win.h roms.h mem.h mon.h kbd.h tape.h joy.h snd.h ctc.h sdxfdc.h sid.h printer.h cpm.h dis.h ui.h vdeb.h memu.h

memu_cfg.o:	memu.c Z80.h types.h common.h diag.h win.h roms.h mem.h mon.h kbd.h tape.h joy.h snd.h ctc.h sdxfdc.h sid.h printer.h cpm.h dis.h ui.h vdeb.h memu.h
		@echo "  CC    $@"
		@$(CC) -c -o $@ $(CFLAGS) -DALT_OPTIONS=cfg_options -DALT_USAGE=cfg_usage -DALT_ENTRY=memu $<

TxtRead.o:	TxtRead.c TxtRead.h cbool.h

clean:
		rm *.o
