#
# Memotech Emulator
#
CC =		cl
LD =		link
TH =		../../th
DIRT =		../../dirt
IDIR =		../../memu
CFLAGS =	/nologo /c  /W3 /WX /MT /Zi /arch:IA32 /I$(TH) /I$(DIRT) /IWin32 /D_CRT_SECURE_NO_DEPRECATE /DWIN32 /DLSB_FIRST /DDEBUG /DBEMEMU /DNO_DART /DHAVE_CFX2 /DHAVE_VGA
LFLAGS =	/NOLOGO /STACK:0x30000 /INCREMENTAL:NO /SUBSYSTEM:CONSOLE,5.01 /DEBUG
LIBS =		Win32/portaudio_x86.lib setargv.obj user32.lib gdi32.lib winmm.lib dinput8.lib

.SUFFIXES:	.c .obj

.c.obj:
		@echo CC    $@
		@$(CC) /c /Fo$@ $(CFLAGS) $*.c

#
all:		$(IDIR)/memu.exe $(IDIR)/memu-cfg.exe

$(IDIR)/memu.exe:		Z80.obj diag.obj common.obj win_w32.obj roms.obj mem.obj vid.obj kbd2.obj tape.obj joy_w32.obj snd.obj ctc.obj monprom.obj mon.obj sdxfdc.obj sid.obj printer.obj spec.obj cpm.obj dis.obj ui.obj cfx2.obj vga.obj vdeb.obj memu.obj $(TH)/th.lib $(DIRT)/dirt.lib
		@echo LD    $@
		@$(LD) /OUT:$@ $(LFLAGS) $** $(LIBS)

$(IDIR)/memu-cfg.exe:	Z80.obj diag.obj common.obj win_w32.obj roms.obj mem_cfg.obj vid.obj kbd_cfg.obj tape.obj joy_w32.obj snd.obj ctc.obj monprom.obj mon.obj sdxfdc.obj sid.obj printer.obj spec.obj cpm.obj dis.obj ui.obj cfx2.obj vga.obj vdeb.obj config.obj memu_cfg.obj $(TH)/th.lib $(DIRT)/dirt.lib
		@echo LD    $@
		@$(LD) /OUT:$@ $(LFLAGS) $** $(LIBS)

Z80.obj:		Z80.c Z80.h Tables.h CodesCB.h CodesXCB.h CodesED.h CodesXX.h Codes.h

diag.obj:		diag.c types.h diag.h vdeb.h

common.obj:	common.c types.h diag.h common.h mem.h vid.h kbd.h tape.h joy.h mon.h sdxfdc.h sid.h printer.h ui.h vdeb.h

win_w32.obj:		win_w32.c types.h common.h win.h

roms.obj:		roms.c types.h roms.h

mem.obj:		mem.c types.h common.h diag.h roms.h mem.h

mem_cfg.obj:	mem.c types.h common.h diag.h roms.h mem.h
		@echo CC    $@
		@$(CC) /c /Fo$@ $(CFLAGS) -DDYNAMIC_ROMS mem.c

vid.obj:		vid.c types.h common.h diag.h mem.h vid.h kbd.h mon.h

kbd2.obj:		kbd2.c types.h common.h diag.h win.h kbd.h vdeb.h

kbd_cfg.obj:	kbd2.c types.h common.h diag.h win.h kbd.h
		@echo CC    $@
		@$(CC) /c /Fo$@ $(CFLAGS) -DALT_KEYPRESS=test_cfg_key kbd2.c

tape.obj:		tape.c types.h memu.h tape.h

joy_w32.obj:		joy_w32.c types.h common.h diag.h kbd.h joy.h

snd.obj:		snd.c types.h common.h diag.h snd.h

ctc.obj:		ctc.c types.h common.h diag.h ctc.h

monprom.obj:	monprom.c types.h monprom.h

mon.obj:		mon.c types.h common.h diag.h win.h kbd.h monprom.h mon.h $(TH)/th.h

sdxfdc.obj:	sdxfdc.c types.h diag.h sdxfdc.h

sid.obj:		sid.c types.h diag.h common.h sid.h

printer.obj:	printer.c types.h common.h diag.h printer.h

spec.obj:		spec.c types.h diag.h common.h spec.h

cpm.obj:		cpm.c Z80.h types.h common.h diag.h mem.h mon.h printer.h $(DIRT)/dirt.h

dis.obj:		dis.c types.h mem.h dis.h

ui.obj:		ui.c types.h common.h win.h mem.h vid.h monprom.h dis.h ui.h

cfx2.obj:		cfx2.c types.h common.h diag.h cfx2.h

vga.obj:		vga.c vga.h monprom.h win.h vid.h kbd.h common.h diag.h

vdeb.obj:		vdeb.c vdeb.h memu.h mem.h dis.h kbd.h monprom.h win.h Z80.h

config.obj:	config.c config.h types.h memu.h monprom.h win.h kbd.h memu.h monprom.h kbd.h mem.h vid.h mon.h snd.h sdxfdc.h common.h
		@echo "  CC    $@"
		@$(CC) /c /Fo$@ $(CFLAGS) -DALT_MAIN config.c

gpio.obj:		gpio.c gpio.h

rpi.obj:		rpi.c config.h types.h common.h gpio.h

memu.obj:		memu.c Z80.h types.h common.h diag.h win.h roms.h mem.h mon.h kbd.h tape.h joy.h snd.h ctc.h sdxfdc.h sid.h printer.h cpm.h dis.h ui.h vdeb.h memu.h

memu_cfg.obj:	memu.c Z80.h types.h common.h diag.h win.h roms.h mem.h mon.h kbd.h tape.h joy.h snd.h ctc.h sdxfdc.h sid.h printer.h cpm.h dis.h ui.h vdeb.h memu.h
		@echo CC    $@
		@$(CC) /c /Fo$@ $(CFLAGS) -DALT_OPTIONS=cfg_options -DALT_USAGE=cfg_usage -DALT_ENTRY=memu memu.c

hardware.obj:	hardware.c hardware.h types.h gpio.h TxtRead.h cbool.h

TxtRead.obj:	TxtRead.c TxtRead.h cbool.h

clean:
		del *.obj
