/*

dirt.c - Directory Traversal - NetWare version

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <dos.h>
#include <direct.h>

#define	DIRT_C
#include "dirt.h"

/*...vdirt\46\h:0:*/
/*...e*/

typedef int BOOLEAN;
#define	TRUE  1
#define	FALSE 0

typedef struct
	{
#ifdef DIRT_USE_OPENDIR
	DIR *dir;
#else
	struct find_t find;
	BOOLEAN empty, first;
#endif
	} DIRT;

/*...sdirt_open:0:*/
#define	L_FILENAME (2+64+8+1+3)		/* d:\path\filename.ext */

DIRT *dirt_open(const char *dirname, int *rc)
	{
#ifdef DIRT_USE_OPENDIR
	*rc = DIRTE_OK;
	if ( (dirt = (DIRT *) malloc(sizeof(DIRT))) == NULL )
		{
		*rc = DIRTE_NO_MEMORY;
		return NULL;
		}
	if ( (dirt->dir = opendir(dirname)) == NULL )
		{
		switch ( errno )
			{
			case ENOMEM:	*rc = DIRTE_NO_MEMORY;		break;
			case ENOENT:	*rc = DIRTE_NOT_FOUND;		break;
			case ENOTDIR:	*rc = DIRTE_NOT_DIRECTORY;	break;
			case EACCES:	*rc = DIRTE_NO_ACCESS;		break;
			case EMFILE:
			case ENFILE:	*rc = DIRTE_NO_RESOURCE;	break;
			default:	*rc = DIRTE_GEN_ERROR;		break;
			}
		free(dirt);
		return NULL;
		}
	return dirt;
#else	
	DIRT *dirt;
	char pattern[L_FILENAME+4+1];
	unsigned drc;
	int len = strlen(dirname);
	*rc = DIRTE_OK;
	if ( len == 0 || len > L_FILENAME )
		{
		*rc = DIRTE_NOT_FOUND;
		return NULL;
		}
	if ( (dirt = (DIRT *) malloc(sizeof(DIRT))) == NULL )
		{
		*rc = DIRTE_NO_MEMORY;
		return NULL;
		}
	memcpy(pattern, dirname, len);
	if ( pattern[len-1] == '/' || pattern[len-1] == '\\' )
		strcpy(pattern+len-1, "\\*");
	else if ( len == 2 && isalpha(pattern[0]) && pattern[1] == ':' )
		strcpy(pattern+len, "*");
	else
		strcpy(pattern+len, "\\*");
	drc = DOSFindFirstFile(pattern, _A_HIDDEN|_A_SYSTEM|_A_SUBDIR, &(dirt->find));
	if ( drc != 0 )
		{
		switch ( drc )
			{
			case 0x02: /* file not found */
			case 0x03: /* path not found */
				*rc = DIRTE_NOT_FOUND;
				break;
			case 0x12: /* no more files */
				dirt->empty = TRUE;
				return dirt;
			default:
				*rc = DIRTE_GEN_ERROR;
				break;
			}
		free(dirt);
		return NULL;
		}
	dirt->empty = FALSE;
	dirt->first = TRUE;
	return dirt;
#endif
	}
/*...e*/
/*...sdirt_next:0:*/
const char *dirt_next(DIRT *dirt)
	{
#ifdef DIRT_USE_OPENDIR
	DIR *d;
	if ( (d = readdir(dirt->dir)) == NULL )
		return NULL;
	return d->d_name;
#else
	if ( dirt->empty )
		return NULL;
	else if ( dirt->first )
		dirt->first = FALSE;
	else
		{
		unsigned drc = DOSFindNextFile(&(dirt->find));
		if ( drc != 0 )
			return NULL;
		}
	return dirt->find.name;
#endif
	}
/*...e*/
/*...sdirt_close:0:*/
void dirt_close(DIRT *dirt)
	{
#ifdef DIRT_USE_OPENDIR
	closedir(dirt->dir);
#endif
	free(dirt);
	}
/*...e*/

/*...sdirt_error:0:*/
const char *dirt_error(int rc)
	{
	const char *errlist[] =
		{
		NULL,
		"general error",
		"out of memory",
		"not found",
#ifdef DIRT_USE_OPENDIR
		"not a directory",
		"access denied",
		"out of resource",
#endif
		};
	return errlist[rc];
	}
/*...e*/
