/*

th.c - Terminal Handler (Amiga version)

Large amounts of inspiration (pirated code to you) obtained from Emacs 3.8

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <dos.h>
#define	TH_C
#include "th.h"

/*...vth\46\h:0:*/
/*...e*/

/*...sforward references:0:*/
void ttputc(), amg_flush(), th_p_raw_str();
/*...e*/
/*...svars:0:*/
#define	ESC	0x1b
#define	CSI	0x9b
#define	BEL	0x07

int x_size;
int y_size;

static int csr_type = CSR_OFF;		/* What sort of cursor do we have    */
static int csr_x, csr_y;
/*...e*/
/*...sansiparm:0:*/
static void ansiparm(register int n)
	{
	register int q, r;

	if ( (q = n / 10) )
		{
		if ( (r = q / 10) )
			ttputc((r % 10) + '0');
		ttputc((q % 10) + '0');
		}
	ttputc((n % 10) + '0');
	}
/*...e*/
/*...samiga i\47\o:0:*/
#define NEW 1006
#define AMG_MAXBUF	1024L
static long terminal;
static char scrn_tmp [AMG_MAXBUF+1];
static long scrn_tmp_p = 0;

static void ttputc(char c)
	{
	scrn_tmp [scrn_tmp_p++] = c;
	if( scrn_tmp_p >= AMG_MAXBUF )
		amg_flush();
	}

static void amg_flush(void)
	{
	if( scrn_tmp_p )
		Write(terminal, scrn_tmp, scrn_tmp_p);
	scrn_tmp_p = 0;
	}

static int ttgetc(void)
	{
	char ch;
	amg_flush();
	Read(terminal, &ch, 1L);
	return(255 & (int) ch);
	}
/*...e*/
/*...swipe_bottom_line:0:*/
static void wipe_bottom_line(void)
	{
	int x = csr_x, y = csr_y;
	low_level_cxy(0, y_size);
	tputs(cap_ce, 1, outc);
	low_level_cxy(x, y);
	}
/*...e*/

/*...sth_beep:0:*/
void th_beep(void)
	{
	ttputc(BEL);
	}
/*...e*/
/*...sth_cxy:0:*/
static void cxy2(int x, int y)
	{
	ttputc(CSI);
	ansiparm(y + 1);
	ttputc(';');
	ansiparm(x + 1);
	ttputc('H');
	csr_x = x;
	csr_y = y;
	}

void th_cxy(int x, int y)
	{
	cxy2(x, y);
	amg_flush();
	}
/*...e*/
/*...sth_cy:0:*/
void th_cy(int y)
	{
	cxy2(0, y);
	amg_flush();
	}
/*...e*/
/*...sth_chome:0:*/
void th_chome(void)
	{
	cxy2(0, 0);
	amg_flush();
	}
/*...e*/
/*...sth_cls:0:*/
void th_cls(void)
	{
	ttputc(CSI);
	ttputc('2');
	ttputc('J');
	amg_flush();
	}
/*...e*/
/*...sth_setcol:0:*/
void th_setcol(int col)
	{
	ttputc(CSI);
	ansiparm(((col>>3) & 0x7) + 30);
	ttputc('m');
	ttputc(CSI);
	ansiparm((col & 0x7) + 40);
	ttputc('m');
	amg_flush();
	}
/*...e*/
/*...sth_setcsr:0:*/
void th_setcsr(int csr)
	{
	ttputc(CSI);
	if ( (csr_type = csr) == CSR_OFF )
		ttputc(0x30);
	ttputc(0x20);
	ttputc(0x70);
	amg_flush();
	}
/*...e*/
/*...sth_inslin:0:*/
void th_inslin(void)
	{
	th_cy(csr_y);
	ttputc(CSI);
	ttputc('1');
	ttputc('L');
	wipe_bottom_line();
	amg_flush();
	}
/*...e*/
/*...sth_dellin:0:*/
void th_dellin(void)
	{
	th_cxy(0, csr_y);
	ttputc(CSI);
	ttputc('1');
	ttputc('M');
	wipe_bottom_line();
	amg_flush();
	}
/*...e*/
/*...sth_deleol:0:*/
void th_deleol(void)
	{
	ttputc(CSI);
	ttputc('K');
	amg_flush();
	}
/*...e*/
/*...sth_isprint:0:*/
/* Conservative estimate here (I don't actually know whats really safe!) */

int th_isprint(int c)
	{
	return c >= ' ' && c <= '~';
	}
/*...e*/
/*...sth_p_raw_chr:0:*/
static void low_level_p_raw_chr(register int c)
	{
	ttputc(c);
	if ( ++csr_x == x_size )
		{
		csr_x = 0; csr_y++;
		}
	}

void th_p_raw_chr(register int c)
	{
	low_level_p_raw_chr(c);
	amg_flush();
	}
/*...e*/
/*...sth_p_raw_str:0:*/
void th_p_raw_str(const char *s)
	{
	while ( *s )
		low_level_p_raw_chr(*s++);
	amg_flush();
	}
/*...e*/
/*...sth_p_chr:0:*/
static void low_level_p_chr(register int c)
	{
	if ( c == '\n' )
		cxy2(csr_x = 0, ++csr_y);
	else if ( c == '\t' )
		{
		if ( (csr_x = ((csr_x+8) & 0xf8)) == x_size )
			{
			csr_x = 0; csr_y++;
			}
		cxy2(csr_x, csr_y);
		}
	else if ( c == '\b' )
		{
		if ( csr_x )
			csr_x--;
		else
			{
			csr_x = x_size - 1;
			csr_y--;
			}
		cxy2(csr_x, csr_y);
		}
	else if ( c == '\r' )
		cxy2(csr_x = 0, csr_y);
	else
		low_level_p_raw_chr(c);
	}

void th_p_chr(register int c)
	{
	low_level_p_chr(c);
	amg_flush();
	}
/*...e*/
/*...sth_p_str:0:*/
void th_p_str(register const char *s)
	{
	while ( *s )
		low_level_p_chr(*s++);
	amg_flush();
	}
/*...e*/
/*...sth_sync_screen:0:*/
void th_sync_screen(void)
	{
	}
/*...e*/
/*...skeyboard:0:*/
/*
This bit is a little convoluted. It might even work!
*/

int th_key_read(void)
	{
	register int c;
	register int d;

	if ( (c = ttgetc()) == CSI )
		/* Is a function key or shifted function key */
		/* or an arrow key or the help key */
		{
		switch ( c = ttgetc() )
			{
			case 'A':	return ( K_UP );
			case 'B':	return ( K_DOWN );
			case 'C':	return ( K_RIGHT );
			case 'D':	return ( K_LEFT );
			case 'S':	return ( K_SDOWN );
			case 'T':	return ( K_SUP );
			case ' ':	return ( ttgetc() == 'A' ?
							K_SLEFT : K_SRIGHT );
			case '?':	ttgetc();
					return ( K_HELP );
			}

		/* => function or shifted function */

		if ( (d = ttgetc()) == '~' )
			/* Just a normal function key */
			return ( K_F(c - '0' + 1) );

		/* => must be shifted function key */
		ttgetc();	/* Swallow ~ */
		return ( K_SF(d - '0' + 1) );
		}

	/* => could be normal key or */
	/* Alt-letter or Alt-number or unrecognised */
	if ( c >= 0x80 || c < 0 )
		return ( 0x0300 | c );

	/* Normal key */
	return ( c );
	}

int th_key_status(void) { return 0; }

TH_BOOLEAN th_key_break(void) { return TH_FALSE; }
/*...e*/
/*...sth_init and th_deinit:0:*/
static TH_BOOLEAN th_inited = TH_FALSE;

TH_BOOLEAN th_init(int th_mode, int desired_x_size, int desired_y_size)
	{
	char s [100+1];

	x_size = ( desired_x_size != 0 ) ? desired_x_size : 77;
	y_size = ( desired_y_size != 0 ) ? desired_y_size : 22;

	if ( x_size > 77 || y_size > 22 )
		return TH_FALSE;

	sprintf(s, "RAW:1/1/%d/%d/Terminal Handler",
		x_size * 8 + 24 - 1,
		y_size * 8 + 24 - 1);

	terminal = Open(s, NEW);

	th_setcsr(CSR_BLOCK);
	th_setcol(COL(COL_1, COL_0));
	th_cls();
	return th_inited = TH_TRUE;
	}

void th_deinit(void)
	{
	if ( !th_inited )
		return;

	amg_flush();
	Close(terminal);
	}
/*...e*/
