/*

th.h - Interface to Terminal Handler (Amiga version)

*/

#ifndef TH_H
#define	TH_H

typedef int TH_BOOLEAN;
#define	TH_TRUE  1
#define	TH_FALSE 0

#define	CSR_BLOCK	0		/* These have no effect */
#define	CSR_ULINE	1
#define	CSR_ALT		2
#define	CSR_OFF		3

#define	COL_0		0		/* Colours as used by intuition */
#define	COL_1		1
#define	COL_2		2
#define	COL_3		3

#define	COL(f,b)	((f)*0x08+(b))

#define	MODE_NORMAL	0

#define	K_BS		0x0008
#define	K_TAB		0x0009
#define	K_CR		0x000d
#define	K_LF		0x000a
#define	K_ESC		0x001b
#define	K_DEL		0x007f
#define	K_UP		0x0100
#define	K_DOWN		0x0101
#define	K_LEFT		0x0102
#define	K_RIGHT		0x0103
#define	K_SUP		0x0104
#define	K_SDOWN		0x0105
#define	K_SLEFT		0x0106
#define	K_SRIGHT	0x0107
#define	K_HELP		0x0108
#define	K_CTRL(c)	((c)&0x1f)
#define	K_F(n)		(0x0200+(n)-1)
#define	K_SF(n)		(0x0210+(n)-1)

#ifndef TH_C

extern int x_size, y_size;

extern TH_BOOLEAN th_init(int mode, int x_size, int y_size);
extern void th_deinit(void);
extern void th_beep(void);
extern void th_cls(void);
extern void th_cxy(int x, int y);
extern void th_cy(int y);
extern void th_chome(void);
extern void th_setcol(int col);
extern void th_setcsr(int csr);
extern void th_inslin(void);
extern void th_dellin(void);
extern void th_deleol(void);
extern TH_BOOLEAN th_isprint(int c);
extern void th_p_raw_chr(int c);
extern void th_p_raw_str(const char *s);
extern void th_p_chr(int c);
extern void th_p_str(const char *s);
extern void th_sync_screen(void);

extern int th_key_read(void);
extern int th_key_status(void);
extern TH_BOOLEAN th_key_break(void);

#endif

#endif
