/*

th.c - Terminal handler (debug version)

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>
#include <termios.h>
#define	TH_C
#include "th.h"

/*...vth\46\h:0:*/
/*...e*/

/*...svars:0:*/
#define	SCN_W 80
#define	SCN_H 50

int x_size = SCN_W;
int y_size = SCN_H;

static char screen[SCN_H][SCN_W];

static int csr_x = 0;
static int csr_y = 0;
/*...e*/
/*...strace:0:*/
void trace(const char *fmt, ...)
	{
	va_list	vars;
	char s[256+1];
	FILE *fp;
	va_start(vars, fmt);
	vsprintf(s, fmt, vars);
	va_end(vars);
	if ( (fp = fopen("th.log", "a")) != NULL )
		{
		fputs(s, fp);
		fputc('\n', fp);
		fclose(fp);
		}
	}
/*...e*/
/*...sfatal:0:*/
static void fatal(const char *fmt, ...)
	{
	va_list	vars;
	char s[256+1];
	va_start(vars, fmt);
	vsprintf(s, fmt, vars);
	va_end(vars);
	fprintf(stderr, "Violation of TH API: %s\n", s);
	exit(1);
	}
/*...e*/

/*...sth_beep:0:*/
void th_beep(void)
	{
	}
/*...e*/
/*...sth_cxy:0:*/
void th_cxy(int x, int y)
	{
	trace("th_cxy(%d,%d)", x, y);
	if ( x < 0 || x > SCN_W ) fatal("th_cxy, x=%d", x);
	if ( y < 0 || y > SCN_H ) fatal("th_cxy, y=%d", y);
	csr_x = x;
	csr_y = y;
	}
/*...e*/
/*...sth_cy:0:*/
void th_cy(int y)
	{
	trace("th_cy(%d)", y);
	if ( y < 0 || y > SCN_H ) fatal("th_cxy, y=%d", y);
	csr_x = 0;
	csr_y = y;
	}
/*...e*/
/*...sth_chome:0:*/
void th_chome(void)
	{
	trace("th_chome()");
	csr_x = 0;
	csr_y = 0;
	}
/*...e*/
/*...sth_cls:0:*/
void th_cls(void)
	{
	memset(screen, ' ', SCN_H*SCN_W);
	csr_x = 0;
	csr_y = 0;
	}
/*...e*/
/*...sth_setcol:0:*/
void th_setcol(int col)
	{
	}
/*...e*/
/*...sth_setcsr:0:*/
void th_setcsr(int csr)
	{
	}
/*...e*/
/*...sth_inslin:0:*/
void th_inslin(void)
	{
	int y;
	trace("th_inslin()");
	if ( csr_y < 0 || csr_y >= SCN_H ) fatal("th_inslin, csr_y=%d", csr_y);
	for ( y = SCN_H-1; y > csr_y; y-- )
		memcpy(&(screen[y]), &(screen[y-1]), SCN_W);
	memset(&(screen[csr_y]), ' ', SCN_W);
	}
/*...e*/
/*...sth_dellin:0:*/
void th_dellin(void)
	{
	int y;
	trace("th_dellin()");
	if ( csr_y < 0 || csr_y >= SCN_H ) fatal("th_dellin, csr_y=%d", csr_y);
	for ( y = csr_y; y < SCN_H-1; y++ )
		memcpy(&(screen[y]), &(screen[y+1]), SCN_W);
	memset(&(screen[SCN_H-1]), ' ', SCN_W);
	}
/*...e*/
/*...sth_deleol:0:*/
void th_deleol(void)
	{
	trace("th_deleol()");
	memset(&(screen[csr_y][csr_x]), ' ', SCN_W-csr_x);
	}
/*...e*/
/*...sth_isprint:0:*/
TH_BOOLEAN th_isprint(int c)
	{
	return isprint(c);
	}
/*...e*/
/*...sth_p_raw_chr:0:*/
static void p_raw_chr(int c)
	{
	if ( csr_x < 0 || csr_x >= SCN_W ) fatal("p_raw_chr, csr_x=%d", csr_x);
	if ( csr_y < 0 || csr_y >= SCN_H ) fatal("p_raw_chr, csr_y=%d", csr_y);
	screen[csr_y][csr_x] = (char) c;
	if ( ++csr_x == x_size )
		{
		csr_x = 0; csr_y++;
		}
	}

void th_p_raw_chr(int c)
	{
	trace("th_p_raw_chr('%c')", c);
	p_raw_chr(c);
	}
/*...e*/
/*...sth_p_raw_str:0:*/
static void p_raw_str(const char *s)
	{
	if ( csr_x < 0 || csr_x >= SCN_W ) fatal("p_raw_str, csr_x=%d", csr_x);
	if ( csr_y < 0 || csr_y >= SCN_H ) fatal("p_raw_str, csr_y=%d", csr_y);
	while ( *s )
		{
		if ( csr_y >= SCN_H ) fatal("p_raw_str, csr_y=%d", csr_y);
		screen[csr_y][csr_x] = (char) *s++;
		if ( ++csr_x == x_size )
			{
			csr_x = 0; csr_y++;
			}
		}
	}

void th_p_raw_str(const char *s)
	{
	trace("th_p_raw_str(\"%s\")", s);
	p_raw_str(s);
	}
/*...e*/
/*...sth_p_chr:0:*/
void th_p_chr(int c)
	{
	if ( c == '\n' )
		{
		trace("th_p_chr('\n')");
		csr_x = 0; ++csr_y;
		}
	else if ( c == '\t' )
		{
		trace("th_p_chr('\t')");
		if ( (csr_x = ((csr_x+8) & 0xf8)) == x_size )
			{
			csr_x = 0; csr_y++;
			}
		}
	else if ( c == '\b' )
		{
		trace("th_p_chr('\b')");
		if ( csr_x )
			csr_x--;
		else
			{
			csr_x = x_size - 1;
			csr_y--;
			}
		}
	else if ( c == '\r' )
		{
		trace("th_p_chr('\r')");
		csr_x = 0;
		}
	else
		{
		trace("th_p_chr('%c')", c);
		p_raw_chr(c);
		}
	}
/*...e*/
/*...sth_p_str:0:*/
void th_p_str(const char *s)
	{
	char buf[1024+1];
	int i;
	strcpy(buf, s);
	for ( i = 0; buf[i]; i++ )
		if ( !isprint(buf[i]) )
			buf[i] = '?';
	trace("th_p_str(\"%s\")", buf);

	while ( *s )
		p_raw_chr(*s++);
	}
/*...e*/
/*...sth_sync_screen:0:*/
void th_sync_screen(void)
	{
	int x, y;
	trace("th_sync_screen()");
	fputs("\x1b[H", stdout);
	for ( y = 0; y < SCN_H; y++ )
		{
		for ( x = 0; x < SCN_W; x++ )
			fputc(screen[y][x], stdout);
		fputc('\n', stdout);
		}
	fprintf(stdout, "\x1b[%d;%dH", csr_y+1, csr_x+1);
	fflush(stdout);
	}
/*...e*/
/*...skeyboard stuff:0:*/
int th_key_read(void)
	{
	struct termios oldt, newt;
	int ch;
	tcgetattr(0, &oldt );
	newt = oldt;
	newt.c_lflag &= ~( ICANON | ECHO );
	tcsetattr(0, TCSANOW, &newt );
	ch = getchar();
	tcsetattr(0, TCSANOW, &oldt );
	trace("th_key_read() returns %d", ch);
	return ch;
	}

int th_key_status(void)
	{
	return 0;
	}

TH_BOOLEAN th_key_break(void)
	{
	return TH_FALSE;
	}
/*...e*/
/*...sth_init and th_deinit:0:*/
TH_BOOLEAN th_init(int th_mode, int desired_x_size, int desired_y_size)
	{
	trace("th_init(...)");
	return TH_TRUE;
	}

void th_deinit(void)
	{
	trace("th_deinit()");
	}
/*...e*/
