/*

th.c - Terminal handler (NetWare version)

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <malloc.h>
#ifdef N_PLAT_NLM
#include <nwconio.h>
#else
#include <conio.h>
#endif
#define	TH_C
#include "th.h"

/*...vth\46\h:0:*/

/* Note: In the latest NLM SDK, the symbol N_PLAT_NLM is defined
   in nwtypes.h, which is included by malloc.h. Also, the filename
   of the screen handling functions is nwconio.h, not conio.h. */
/*...e*/
/*...svars:0:*/
int x_size;				/* Width of screen                   */
int y_size;				/* Height of screen                  */

static int csr_x;			/* Current x in range 0 to y_size-1  */
static int csr_y;			/* Current y in range 0 to x_size-1  */
static BYTE csr_col;			/* Current colour                    */

typedef struct { BYTE ch; BYTE attr; } CELL;

#define	MAX_WIDTH 132
static CELL cells[MAX_WIDTH];
/*...e*/

/*...sth_beep:0:*/
void th_beep(void)
	{
	RingTheBell();
	}
/*...e*/
/*...sth_cxy:0:*/
void th_cxy(int x, int y)
	{
	csr_x = x; csr_y = y;
	}
/*...e*/
/*...sth_cy:0:*/
void th_cy(int y)
	{
	csr_x = 0; csr_y = y;
	}
/*...e*/
/*...sth_chome:0:*/
void th_chome(void)
	{
	csr_x = 0; csr_y = 0;
	}
/*...e*/
/*...sth_cls:0:*/
void th_cls(void)
	{
	int i;
	for ( i = 0; i < x_size; i++ )
		{
		cells[i].ch   = ' ';
		cells[i].attr = csr_col;
		}
	for ( i = 0; i < y_size; i++ )
		CopyToScreenMemory(1, x_size, (BYTE *) cells, 0, i);
	csr_x = 0; csr_y = 0;
	}
/*...e*/
/*...sth_setcol:0:*/
void th_setcol(int col)
	{
	csr_col = (BYTE) col;
	}
/*...e*/
/*...sth_setcsr:0:*/
void th_setcsr(int csr)
	{
	BYTE firstline, lastline;
	GetCursorSize(&firstline, &lastline);
	switch ( csr )
		{
		case CSR_BLOCK:
			SetCursorShape(firstline, lastline);
			DisplayInputCursor();
			break;
		case CSR_ULINE:
			SetCursorShape(lastline-2, lastline);
			DisplayInputCursor();
			break;
		case CSR_ALT:
			SetCursorShape((firstline+lastline)/2, lastline);
			DisplayInputCursor();
			break;
		case CSR_OFF:
			HideInputCursor();
			break;
		}
	}
/*...e*/
/*...sth_inslin:0:*/
void th_inslin(void)
	{
	ScrollScreenRegionDown(csr_y, y_size-csr_y);
	}
/*...e*/
/*...sth_dellin:0:*/
void th_dellin(void)
	{
	ScrollScreenRegionUp(csr_y, y_size-csr_y);
	}
/*...e*/
/*...sth_deleol:0:*/
void th_deleol(void)
	{
	int i;
	for ( i = 0; i < x_size-csr_x; i++ )
		{
		cells[i].ch   = ' ';
		cells[i].attr = csr_col;
		}
	CopyToScreenMemory(1, x_size-csr_x, (BYTE *) cells, csr_x, csr_y);
	}
/*...e*/
/*...sth_save:0:*/
void th_save(int x, int y, int n, void *buf)
	{
	CopyFromScreenMemory(1, n, buf, x, y);
	}
/*...e*/
/*...sth_restore:0:*/
void th_restore(int x, int y, int n, const void *buf)
	{
	CopyToScreenMemory(1, n, (void *) buf, x, y);
	}
/*...e*/
/*...sth_isprint:0:*/
/* All characters except NUL are printable. */

TH_BOOLEAN th_isprint(int c)
	{
	return c != 0;
	}
/*...e*/
/*...sth_p_raw_chr:0:*/
void th_p_raw_chr(register int c)
	{
	CELL cell;
	cell.ch   = (BYTE) c;
	cell.attr = csr_col;
	CopyToScreenMemory(1, 1, (BYTE *) &cell, csr_x, csr_y);
	if ( ++csr_x >= x_size )
		/* moved to next line */
		{
		csr_x -= x_size;
		csr_y++;
		}
	gotoxy(csr_x, csr_y);
	}
/*...e*/
/*...sth_p_raw_str:0:*/
void th_p_raw_str(register const char *s)
	{
	int i, len = strlen(s);
	for ( i = 0; i < len; i++ )
		{
		cells[i].ch   = *s++;
		cells[i].attr = csr_col;
		}
	CopyToScreenMemory(1, len, (BYTE *) cells, csr_x, csr_y);
	if ( (csr_x += len) >= x_size )
		/* moved to next line */
		{
		csr_x -= x_size;
		csr_y++;
		}
	}
/*...e*/
/*...sth_p_chr:0:*/
void th_p_chr(register int c)
	{
	if ( c == '\n' )
		{ csr_x = 0; ++csr_y; }
	else if ( c == '\t' )
		{
		if ( (csr_x = ((csr_x+8) & 0xf8)) == x_size )
			{ csr_x = 0; ++csr_y; }
		}
	else if ( c == '\b' )
		{
		if ( csr_x )
			csr_x--;
		else
			{
			csr_x = x_size - 1;
			csr_y--;
			}
		}
	else if ( c == '\r' )
		csr_x = 0;
	else
		th_p_raw_chr(c);
	}
/*...e*/
/*...sth_p_str:0:*/
void th_p_str(register const char *s)
	{
	while ( *s )
		th_p_chr(*s++);
	}
/*...e*/
/*...sth_sync_screen:0:*/
void th_sync_screen(void)
	{
	gotoxy(csr_x, csr_y);
	}
/*...e*/
/*...skeyboard stuff:0:*/
static int k_saved = 0;

/*...sth_key_read:0:*/
int th_key_read(void)
	{
	int k;
	th_sync_screen();
	if ( (k = k_saved) != 0 )
		{
		k_saved = 0;
		return k;
		}
	k = getch();
	if ( k == 0 || k == 0xe0 )
		k = ( getch() << 8 );
	return k;
	}
/*...e*/
/*...sth_key_status:0:*/
int th_key_status(void)
	{
	int k;
	th_sync_screen();
	if ( (k = k_saved) != 0 )
		return k;
	if ( !kbhit() )
		return 0;
	k = getch();
	if ( k == 0 || k == 0xe0 )
		k = ( getch() << 8 );
	k_saved = k;
	return k;
	}
/*...e*/

static TH_BOOLEAN break_occurred = TH_FALSE;

/*...sth_key_break:0:*/
TH_BOOLEAN th_key_break(void)
	{
	if ( !break_occurred )
		return TH_FALSE;

	break_occurred = TH_FALSE;
	return TH_TRUE;
	}
/*...e*/
/*...sctrl_break:0:*/
static void ctrl_break()
	{
	break_occurred = TH_TRUE;
	signal(SIGINT, ctrl_break);	/* Set for next time */
	}
/*...e*/
/*...e*/
/*...sth_init and th_deinit:0:*/
static TH_BOOLEAN th_inited = TH_FALSE;

TH_BOOLEAN th_init(int th_mode, int desired_x_size, int desired_y_size)
	{
	WORD w, h;
	GetSizeOfScreen(&h, &w);
	x_size = w; y_size = h;
	if ( (desired_x_size != 0 && desired_x_size != x_size) ||
	     (desired_y_size != 0 && desired_y_size != y_size) )
		return TH_FALSE;
	if ( x_size > MAX_WIDTH )
		return TH_FALSE;

	SetAutoScreenDestructionMode(TH_TRUE);
	SetCtrlCharCheckMode(TH_FALSE);
	SetCursorCouplingMode(TH_TRUE);

	if ( signal(SIGINT, SIG_IGN) != SIG_IGN )
		signal(SIGINT, ctrl_break);

	th_setcsr(CSR_BLOCK);
	th_setcol(COL(COL_WHITE, COL_BLACK, TH_FALSE, TH_FALSE));
	th_cls();
	th_chome();

	return th_inited = TH_TRUE;
	}

void th_deinit(void)
	{
	if ( !th_inited )
		return;

	th_setcsr(CSR_BLOCK);
	th_setcol(COL(COL_WHITE, COL_BLACK, TH_FALSE, TH_FALSE));
	th_cls();
	th_chome();

	signal(SIGINT, SIG_DFL);

	th_inited = TH_FALSE;
	}
/*...e*/
